/*  costfns.h

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 2001 University of Oxford  */

/*  CCOPYRIGHT  */


#if !defined(__costfns_h)
#define __costfns_h

#include "newimageall.h"

using namespace NEWIMAGE;

namespace NEWIMAGE {

  enum costfns { Woods, CorrRatio, MutualInfo, NormCorr, NormMI, LeastSq, 
		 NormCorrSinc };

  class Costfn {
  public:
    const volume<float> &refvol;
    const volume<float> &testvol;
    const volume<float> &rweight;
    const volume<float> &tweight;
  private:
    int *bindex;
    int no_bins;
    ColumnVector plnp;
    int *jointhist;
    int *marghist1;
    int *marghist2;
    float *fjointhist;
    float *fmarghist1;
    float *fmarghist2;
    mutable int p_count;
    costfns p_costtype;
    bool validweights;
  public: 
    float smoothsize;
    float fuzzyfrac;

  public:
    // Publicly available calls
    Costfn(const volume<float>& refv, const volume<float>& inputv);
    Costfn(const volume<float>& refv, const volume<float>& inputv,
	   const volume<float>& refweight, const volume<float>& inweight);
    ~Costfn();

    void set_costfn(const costfns& costtype) { p_costtype = costtype; }
    void set_no_bins(int n_bins);
    int count() const { return p_count; }

    // General cost function call
    float cost(const Matrix& affmat) const;    // affmat is voxel to voxel
    // in the following, all warps are mm to mm
    float cost(const volume4D<float>& warp) const;
    float cost_gradient(volume4D<float>& gradvec,
			const volume4D<float>& warp, bool nullbc) const; 

    // some basic entropy calls
    float ref_entropy(const Matrix& aff) const;
    float test_entropy(const Matrix& aff) const;
    float joint_entropy(const Matrix& aff) const;

  private:
    // Prevent default behaviours
    Costfn();
    Costfn operator=(const Costfn&);
    Costfn(const Costfn&);

    // Internal functions available
    float normcorr(const Matrix& aff) const; 
    float normcorr_smoothed(const Matrix& aff) const;
    float normcorr_smoothed_sinc(const Matrix& aff) const;
    float normcorr_fully_weighted(const Matrix& aff,
				  const volume<float>& refweight, 
				  const volume<float>& testweight) const;
    
    float leastsquares(const Matrix& aff) const;
    float leastsquares_smoothed(const Matrix& aff) const;
    float leastsquares_fully_weighted(const Matrix& aff, 
				      const volume<float>& refweight, 
				      const volume<float>& testweight) const;
    
    float woods_fn(const Matrix& aff) const; 
    float woods_fn_smoothed(const Matrix& aff) const; 
    
    float corr_ratio(const Matrix& aff) const; 
    float corr_ratio_smoothed(const Matrix& aff) const; 
    float corr_ratio_fully_weighted(const Matrix& aff,
				    const volume<float>& refweight, 
				    const volume<float>& testweight) const;
    float corr_ratio_fully_weighted(const volume4D<float>& warpvol,
				    const volume<float>& refweight, 
				    const volume<float>& testweight) const;
    float corr_ratio_gradient_fully_weighted(volume4D<float>& gradvec,
					     const volume4D<float>& warpvol,
					     const volume<float>& refweight, 
					     const volume<float>& testweight,
					     bool nullbc) const;
     
    float mutual_info(const Matrix& aff) const;
    float mutual_info_smoothed(const Matrix& aff) const;
    float mutual_info_fully_weighted(const Matrix& aff,
				     const volume<float>& refweight, 
				     const volume<float>& testweight) const;
    
    float normalised_mutual_info(const Matrix& aff) const;
    float normalised_mutual_info_smoothed(const Matrix& aff) const;
    float normalised_mutual_info_fully_weighted(const Matrix& aff,
						const volume<float>& refweight, 
						const volume<float>& testweight) const;
    
    float cost(const Matrix& affmat,
	       const volume<float>& refweight, 
	       const volume<float>& testweight) const;
    
    float cost(const volume4D<float>& warp,
	       const volume<float>& refweight, 
	       const volume<float>& testweight) const;
    
    float cost_gradient(volume4D<float>& gradvec,
			const volume4D<float>& warp,
			const volume<float>& refweight, 
			const volume<float>& testweight,
			bool nullbc=false) const;

  };


   //////////////////////////////////////////////////////////////////////////


}

#endif







